@@ -16,7 +16,9 @@ char move[100];
 read_first_move(fd)
 int fd;
 {
-	if (EXP_TIMEOUT == exp_expectl(fd,"first\r\n1.*\r\n",0,(char *)0)) {
+	if (EXP_TIMEOUT == exp_expectl(fd,
+			exp_glob,"first\r\n1.*\r\n",0,
+			exp_end)) {
 		timedout();
 	}
 	sscanf(exp_match,"%*s 1. %s",move);
@@ -27,7 +29,7 @@ int fd;
 read_counter_move(fd)
 int fd;
 {
-	switch (exp_expectl(fd,"*...*\r\n",0,(char *)0)) {
+	switch (exp_expectl(fd,exp_glob,"*...*\r\n",0,exp_end)) {
 	case EXP_TIMEOUT: timedout();
 	case EXP_EOF: exit(-1);
 	}
@@ -38,7 +40,7 @@ int fd;
 read_move(fd)
 int fd;
 {
-	switch (exp_expectl(fd,"*...*\r\n*.*\r\n",0,(char *)0)) {
+	switch (exp_expectl(fd,exp_glob,"*...*\r\n*.*\r\n",0,exp_end)) {
 	case EXP_TIMEOUT: timedout();
 	case EXP_EOF: exit(-1);
 	}
@@ -60,7 +62,7 @@ main(){
 
 	fd1 = exp_spawnl("chess","chess",(char *)0);
 
-	if (-1 == exp_expectl(fd1,"Chess\r\n",0,(char *)0)) exit;
+	if (-1 == exp_expectl(fd1,exp_glob,"Chess\r\n",0,exp_end)) exit;
 
 	if (-1 == write(fd1,"first\r",6)) exit;
 
@@ -68,7 +70,7 @@ main(){
 
 	fd2 = exp_spawnl("chess",0);
 
-	if (-1 == exp_expectl(fd2,"Chess\r\n",0,(char *)0)) exit;
+	if (-1 == exp_expectl(fd2,exp_glob,"Chess\r\n",0,exp_end)) exit;
 
 	for (;;) {
 		send_move(fd2);
