@@ -50,6 +50,8 @@ send_human	5 flts	send -h timing		user
 
 */
 
+#include "exp_conf.h"
+
 #include <stdio.h>
 #include <sys/types.h>
 #include <sys/time.h>
@@ -64,12 +66,11 @@ send_human	5 flts	send -h timing		user
 #ifdef HAVE_STROPTS_H
 #  include <sys/stropts.h>
 #endif
-#ifndef M_XENIX
- /* wait.h is not available on SCO XENIX 386. -- pf@artcom0.north.de */
-#if !defined(CRAY) || CRAY>=60
+
+#ifdef HAVE_SYS_WAIT_H
 #include <sys/wait.h>
 #endif
-#endif
+
 #include <varargs.h>
 #include <errno.h>
 
@@ -77,6 +78,8 @@ send_human	5 flts	send -h timing		user
 #include <math.h>		/* for log/pow computation in send -h */
 #include <ctype.h>		/* all this for ispunct! */
 
+#include "tclInt.h"		/* need OpenFile */
+
 #include "tcl.h"
 #include "string.h"
 #include "exp_rename.h"
@@ -102,7 +105,6 @@ int exp_forked = FALSE;		/* whether we are child process */
 static int sendCD_error = 2;	/* called as send_error */
 static int sendCD_user = 3;	/* called as send_user */
 static int sendCD_proc = 4;	/* called as send or send_spawn */
-static int sendCD_log = 5;	/* called as send_log */
 static int sendCD_tty = 6;	/* called as send_tty */
 
 struct f *fs = 0;		/* process array (indexed by spawn_id's) */
@@ -136,12 +138,12 @@ int opened;		/* check not closed */
 int adjust;		/* adjust buffer sizes */
 char *msg;
 {
-	if (fd >= 0 && fd <= fd_max && (fs[fd].flags & FD_VALID)) {
+	if (fd >= 0 && fd <= fd_max && (fs[fd].valid)) {
 		struct f *f = fs + fd;
 
 		/* following is a little tricky, do not be tempted do the */
 		/* 'usual' boolean simplification */
-		if ((!opened) || !(f->flags & FD_CLOSED)) {
+		if ((!opened) || !f->user_closed) {
 			if ((!adjust) || (TCL_OK == exp_adjust(interp,f)))
 				return f;
 		}
@@ -167,18 +169,75 @@ int pid;
 }
 #endif
 
+/* Tcl needs commands in writable space */
+static char close_cmd[] = "close";
+
+/* prevent an fd from being allocated */
+void
+exp_busy(fd)
+int fd;
+{
+	int x = open("/dev/null",0);
+	if (x != fd) {
+		fcntl(x,F_DUPFD,fd);
+		close(x);
+	}
+}
+
+/* clean up any remnants of a spawned process */
+void
+exp_destroy(fd)
+int fd;
+{
+	if (!fs[fd].sys_closed) close(fd);
+
+	fs[fd].valid = FALSE;	/* retire this slot */
+}
+
+int exp_close_count = 0;
+
 int
 exp_close(interp,fd)
 Tcl_Interp *interp;
 int fd;
 {
+	char *argv[3];
+
 	struct f *f = exp_fd2f(interp,fd,1,0,"close");
 	if (!f) return(TCL_ERROR);
 
-	/* Ignore close errors.  Some systems are really braindamaged and */
-	/* return errors for no evident reason.  Anyway, receiving an error */
-	/* upon pty-close doesn't mean anything anyway as far as I know. */
+	f->user_closed = TRUE;
+
+	if (f->tcl_handle) {
+		int rc;
+
+		argv[0] = close_cmd;
+		argv[1] = f->tcl_handle;
+		argv[2] = 0;
+
+		/* ClientData isn't used by Tcl_CloseCmd so we can afford */
+		/* to pass a 0, otherwise I don't know what we'd do! */
+		rc = Tcl_CloseCmd((ClientData)0,interp,2,argv);
+		f->sys_closed = TRUE;
+
+		free(f->tcl_handle);
+		f->tcl_handle = 0;
+
+		/* make it appear as if process has been waited for */
+		f->sys_waited = TRUE;
+		f->wait = 0;
+
+		if (rc != TCL_OK) return rc;
+	} else {
+		/* Ignore close errors.  Some systems are really odd and */
+		/* return errors for no evident reason.  Anyway, receiving */
+		/* an error upon pty-close doesn't mean anything anyway as */
+		/* far as I know. */
 	close(fd);
+	}
+
+	exp_close_count++;
+
 	if (f->buffer) {
 		free(f->buffer);
 		f->buffer = 0;
@@ -194,12 +253,16 @@ int fd;
 	}
 	f->armed = FALSE;
 
-	if (f->flags & FD_USERWAITED) f->flags = 0;
-	else f->flags |= FD_CLOSED;
+	if (f->user_waited) {
+		exp_destroy(fd);
+	} else {
+		exp_busy(fd);
+	}
+
 	return(TCL_OK);
 }
 
-static int
+static void
 fd_new(fd,pid)
 int fd;
 int pid;
@@ -210,49 +273,40 @@ int pid;
 	/* resize table if nec */
 	if (fd > fd_max) {
 		if (!fs) {	/* no fd's yet allocated */
-			if (!(newfs = (struct f *)
-				malloc(sizeof(struct f)*(fd+1)))) {
-				/* malloc failed */
-				return(TCL_ERROR);
-			}
+			newfs = (struct f *)malloc(sizeof(struct f)*(fd+1));
 			low = 0;
 		} else {		/* enlarge fd table */
-			if (!(newfs = (struct f *)realloc((char *)fs,
-				sizeof(struct f)*(fd+1)))) {
-				/* malloc failed */
-				return(TCL_ERROR);
-			}
+			newfs = (struct f *)realloc((char *)fs,sizeof(struct f)*(fd+1));
 			low = fd_max+1;
 		}
 		fs = newfs;
 		fd_max = fd;
 		for (i = low; i <= fd_max; i++) { /* init new fd entries */
-			struct f *f;
-
-			f = fs+i;
-			f->pid = 0;
-			f->buffer = 0;
-			f->lower = 0;
-			f->flags = 0;	/* not FD_VALID */
-			f->msize = f->size = 0;
-			f->armed = FALSE;
+			fs[i].valid = FALSE;
 		}
 	}
-	/* close down old table entry if nec */
-	fs[fd].pid = pid;		/* overwrite pid */
-	fs[fd].size = 0;		/* forget old data */
-					/* but leave msize and buffer as is! */
-	if (fs[fd].buffer) fs[fd].buffer[0] = '\0';
 
+	/* this could happen if user does "spawn -open stdin" I suppose */
+	if (fs[fd].valid) return;
+
+	/* close down old table entry if nec */
+	fs[fd].pid = pid;
+	fs[fd].size = 0;
+	fs[fd].msize = 0;
+	fs[fd].buffer = 0;
 	fs[fd].printed = 0;
 	fs[fd].echoed = 0;
 	fs[fd].parity = exp_default_parity;
 	fs[fd].key = expect_key++;
 	fs[fd].force_read = FALSE;
 	fs[fd].armed = FALSE;
+	fs[fd].tcl_handle = 0;
 	fs[fd].umsize = exp_default_match_max;
-	fs[fd].flags = FD_VALID;	/* reset all flags */
-	return(TCL_OK);
+	fs[fd].valid = TRUE;
+	fs[fd].user_closed = FALSE;
+	fs[fd].sys_closed = FALSE;
+	fs[fd].user_waited = FALSE;
+	fs[fd].sys_waited = FALSE;
 }
 
 void
@@ -265,9 +319,9 @@ Tcl_Interp *interp;
 	/* (at least in theory anyway) be later re-opened on a different */
 	/* fd, while stdin might have been redirected away from /dev/tty */
 
-	if (dev_tty != -1) {
+	if (exp_dev_tty != -1) {
 		char dev_tty_str[10];
-		sprintf(dev_tty_str,"%d",dev_tty);
+		sprintf(dev_tty_str,"%d",exp_dev_tty);
 		Tcl_SetVar(interp,"tty_spawn_id",dev_tty_str,0);
 	}
 }
@@ -276,17 +330,25 @@ void
 exp_init_spawn_ids(interp)
 Tcl_Interp *interp;
 {
-	fd_new(0,getpid());
-	fs[0].flags |= FD_USER;
+	fd_new(0,EXP_NOPID);
+	fd_new(1,EXP_NOPID);
+	fd_new(2,EXP_NOPID);
 
-	if (dev_tty != -1) {
-		fd_new(dev_tty,getpid());
+	if (exp_dev_tty != -1) {
+		fd_new(exp_dev_tty,getpid());
 	}
 
 	/* really should be in interpreter() but silly to do on every call */
 	exp_adjust(interp,&fs[0]);
 }
 
+void
+exp_close_on_exec(fd)
+int fd;
+{
+	(void) fcntl(fd,F_SETFD,1);
+}
+
 /* arguments are passed verbatim to execvp() */
 /*ARGSUSED*/
 static int
@@ -299,6 +361,8 @@ char **argv;
 	int slave;
 	int pid;
 	char **a;
+	/* tell Saber to ignore non-use of ttyfd */
+	/*SUPPRESS 591*/
 	int ttyfd;
 	int master;
 	int ttyinit = TRUE;
@@ -306,10 +370,17 @@ char **argv;
 	int echo = TRUE;
 	int console = FALSE;
 	char *argv0 = argv[0];
+	char *openarg = 0;
+	OpenFile *filePtr;
+
+	char buf[4];		/* enough space for a string literal */
+				/* representing a file descriptor */
+#if TCL_MAJOR_VERSION != 6
+	Tcl_DString dstring;
+#endif
 
 	argc--; argv++;
 
-	/* new */
 	for (;argc>0;argc--,argv++) {
 		if (streq(*argv,"-nottyinit")) {
 			ttyinit = FALSE;
@@ -319,14 +390,23 @@ char **argv;
 			echo = FALSE;
 		} else if (streq(*argv,"-console")) {
 			console = TRUE;
+		} else if (streq(*argv,"-open")) {
+			openarg = argv[1];
+			argc--; argv++;
 		} else break;
 	}
 
-	if (argc == 0) {
+	if (openarg && (argc != 0)) {
+		exp_error(interp,"usage: -open [fileXX]\n");
+		return TCL_ERROR;
+	}
+
+	if (!openarg && (argc == 0)) {
 		exp_error(interp,"usage: spawn program [args]");
 		return(TCL_ERROR);
 	}
 
+	if (!openarg) {
 	if (echo) {
 		Log(0,"%s ",argv0);
 		for (a = argv;*a;a++) {
@@ -339,12 +419,58 @@ char **argv;
 		exp_error(interp,"too many programs spawned? - out of ptys");
 		return(TCL_ERROR);
 	}
+	} else {
+		if (echo) Log(0,"%s [open ...]\r\n",argv0);
+
+		if (TclGetOpenFile(interp, openarg, &filePtr) != TCL_OK) {
+			return TCL_ERROR;
+		}
+		master = fileno(filePtr->f);
+	}
 
 	/* much easier to set this, than remember all masters */
 	fcntl(master,F_SETFD,1);	/* close on exec */
 
+	if (openarg) {
+		struct f *f;
+
+		fd_new(master,EXP_NOPID);
+
+		f = exp_fd2f(interp,master,0,0,(char *)0);
+		f->tcl_handle = ckalloc(strlen(openarg)+1);
+
+		/* save file# handle */
+		strcpy(f->tcl_handle,openarg);
+		/* save fd handle for output */
+		if (filePtr->f2) {
+			f->tcl_output = fileno(filePtr->f2);
+			fcntl(master,F_SETFD,1);	/* close on exec */
+		} else {
+			f->tcl_output = master;
+		}
+
+
+		f->tcl_output = (filePtr->f2 ? fileno(filePtr->f2) : master);
+
+		/* tell user id of new process */
+		sprintf(buf,"%d",master);
+		Tcl_SetVar(interp,SPAWN_ID_VARNAME,buf,0);
+
+		sprintf(interp->result,"%d",EXP_NOPID);
+		debuglog("spawn: returns {%s}\r\n",interp->result);
+
+		return TCL_OK;
+	}
+
+#if TCL_MAJOR_VERSION == 6
 	if (NULL == (argv[0] = Tcl_TildeSubst(interp,argv[0])))
 		return TCL_ERROR;
+#else
+	if (NULL == (argv[0] = Tcl_TildeSubst(interp,argv[0],dstring))) {
+		Tcl_DStringFree(dstring);
+		return TCL_ERROR;
+	}
+#endif
 
 	if ((pid = fork()) == -1) {
 		exp_error(interp,"fork: %s",sys_errlist[errno]);
@@ -353,9 +479,11 @@ char **argv;
 
 	if (pid) {
 		/* parent */
-		char buf[3];
+		fd_new(master,pid);
 
-		if (TCL_ERROR == fd_new(master,pid)) return(TCL_ERROR);
+#if TCL_MAJOR_VERSION != 6
+		Tcl_DStringFree(dstring);
+#endif
 
 #ifdef CRAY
 		setptypid(pid);
@@ -390,16 +518,18 @@ char **argv;
 
 	/* child process - do not return from here!  all errors must exit() */
 
-	if (dev_tty != -1) {
-		close(dev_tty);
-		dev_tty = -1;
+	if (exp_dev_tty != -1) {
+		close(exp_dev_tty);
+		exp_dev_tty = -1;
 	}
 
 #ifdef CRAY
 	(void) close(master);
 #endif
 
-#ifdef POSIX
+/* ultrix (at least 4.1-2) fails to obtain controlling tty if setsid */
+/* is called.  setpgrp works though.  */
+#if defined(POSIX) && !defined(ultrix)
 #define DO_SETSID
 #endif
 #ifdef __convex__
@@ -448,7 +578,7 @@ char **argv;
 		exit(-1);
 	}
 
-#if defined(TIOCSCTTY) && !defined(CIBAUD) && !defined(sun)
+#if defined(TIOCSCTTY) && !defined(CIBAUD) && !defined(sun) && !defined(hp9000s300)
 	/* 4.3+BSD way to acquire controlling terminal */
 	/* according to Stevens - Adv. Prog..., p 642 */
 	if (ioctl(0,TIOCSCTTY,(char *)0) < 0) {
@@ -530,6 +660,7 @@ char **argv;
 	close(2);
 	fcntl(0,F_DUPFD,2);	/* duplicate 0 onto 2 */
 
+#if 0
 	/* avoid fflush of cmdfile since this screws up the parents seek ptr */
 	/* There is no portable way to fclose a shared read-stream!!!! */
 	if (exp_cmdfile && (exp_cmdfile != stdin))
@@ -538,6 +669,7 @@ char **argv;
 	if (debugfile) (void) fclose(debugfile);
 	/* (possibly multiple) masters are closed automatically due to */
 	/* earlier fcntl(,,CLOSE_ON_EXEC); */
+#endif
 
         (void) execvp(argv[0],argv);
 	/* Unfortunately, by now we've closed fd's to stderr, logfile and
@@ -546,11 +678,48 @@ char **argv;
 	   part of the program output.  This will be picked up in an
 	   expect or interact command.
 	*/
-	errorlog("execvp(%s): %s\r\n",argv[0],sys_errlist[errno]);
+	errorlog("%s: %s\r\n",argv[0],sys_errlist[errno]);
 	exit(-1);
 	/*NOTREACHED*/
 }
 
+/*ARGSUSED*/
+static int
+cmdExpPid(clientData,interp,argc,argv)
+ClientData clientData;
+Tcl_Interp *interp;
+int argc;
+char **argv;
+{
+	struct f *f;
+	int m = -1;
+
+	int argc_orig = argc;
+	char **argv_orig = argv;
+
+	argc--; argv++;
+
+	for (;argc>0;argc--,argv++) {
+		if (streq(*argv,"-i")) {
+			argc--; argv++;
+			if (*argv) {
+				exp_error(interp,"usage: -i spawn_id");
+				return TCL_ERROR;
+			}
+			m = atoi(*argv);
+		} else break;
+	}
+
+	if (m == -1) {
+		if (exp_update_master(interp,&m,0,0) == 0) return TCL_ERROR;
+	}
+
+	if (0 == (f = exp_fd2f(interp,m,1,0,"pid"))) return TCL_ERROR;
+
+	sprintf(interp->result,"%d",f->pid);
+	return TCL_OK;
+}
+
 /*ARGSUSED*/
 static int
 cmdGetpid(clientData, interp, argc, argv)
@@ -593,33 +762,43 @@ char **argv;
 	int total_len = 0, arg_len;
 
 	int stty_args_recognized = TRUE;
+	int stty_cmd = FALSE;
 	int cooked = FALSE;
+	int was_raw, was_echo;
 
-	if (argc > 2 && streq(argv[1],"stty")) {
-		ioctled_devtty = TRUE;
+	if (argc == 1) return TCL_OK;
+
+	if (streq(argv[1],"stty")) {
+		stty_cmd = TRUE;
+		was_raw = exp_israw();
+		was_echo = exp_isecho();
+	}
+
+	if (argc > 2 && stty_cmd) {
+		exp_ioctled_devtty = TRUE;
 
 		for (i=2;i<argc;i++) {
 			if (streq(argv[i],"raw") ||
 			    streq(argv[i],"-cooked")) {
-				tty_raw(1);
+				exp_tty_raw(1);
 			} else if (streq(argv[i],"-raw") ||
 				   streq(argv[i],"cooked")) {
 				cooked = TRUE;
-				tty_raw(-1);
+				exp_tty_raw(-1);
 			} else if (streq(argv[i],"echo")) {
-				tty_echo(1);
+				exp_tty_echo(1);
 			} else if (streq(argv[i],"-echo")) {
-				tty_echo(-1);
+				exp_tty_echo(-1);
 			} else stty_args_recognized = FALSE;
 		}
 		/* if unknown args, fall thru and let real stty have a go */
 		if (stty_args_recognized) {
 #ifdef POSIX
- 			if (tcsetattr(dev_tty,TCSADRAIN, &tty_current) == -1) {
+ 			if (tcsetattr(exp_dev_tty,TCSADRAIN, &tty_current) == -1) {
 #else
-		        if (ioctl(dev_tty, TCSETSW, &tty_current) == -1) {
+		        if (ioctl(exp_dev_tty, TCSETSW, &tty_current) == -1) {
 #endif
-			    if (exp_disconnected || (dev_tty == -1) || !isatty(dev_tty)) {
+			    if (exp_disconnected || (exp_dev_tty == -1) || !isatty(exp_dev_tty)) {
 				errorlog("system stty: impossible in this context\n");
 				errorlog("are you disconnected or in a batch, at, or cron script?");
 				/* user could've conceivably closed /dev/tty as well */
@@ -627,6 +806,11 @@ char **argv;
 			    exp_error("system stty: ioctl(user): %s\r\n",sys_errlist[errno]);
 			    return(TCL_ERROR);
 			}
+			if (stty_cmd) {
+				sprintf(interp->result,"%sraw %secho",
+					(was_raw?"":"-"),
+					(was_echo?"":"-"));
+			}
 			return(TCL_OK);
 		}
 	}
@@ -652,9 +836,9 @@ char **argv;
 	if (!stty_args_recognized) {
 		/* find out what weird options user asked for */
 #ifdef POSIX
- 			if (tcgetattr(dev_tty, &tty_current) == -1) {
+		if (tcgetattr(exp_dev_tty, &tty_current) == -1) {
 #else
-	        if (ioctl(dev_tty, TCGETS, &tty_current) == -1) {
+	        if (ioctl(exp_dev_tty, TCGETS, &tty_current) == -1) {
 #endif
 			errorlog("ioctl(get): %s\r\n",sys_errlist[errno]);
 			exp_exit(interp,-1);
@@ -665,6 +849,12 @@ char **argv;
 		}
 	}
 
+	if (stty_cmd) {
+		sprintf(interp->result,"%sraw %secho",
+			(was_raw?"":"-"),
+			(was_echo?"":"-"));
+	}
+
 	/* Following the style of Tcl_ExecCmd, we can just return the */
 	/* raw result (appropriately shifted and masked) to Tcl */
 	return(0xff & (i>>8));
@@ -842,9 +1032,13 @@ struct human_arg *arg;
 		else alpha = arg->alpha;
 		in_word = !(ispunct(*sp) || isspace(*sp));
 
-		t = (pow(-log((double)unit_random()),arg->c)*alpha)+arg->min;
-		if (t>arg->max) t = arg->max;
+		t = alpha * pow(-log((double)unit_random()),arg->c);
 
+		/* enforce min and max times */
+		if (t<arg->min) t = arg->min;
+		else if (t>arg->max) t = arg->max;
+
+/*fprintf(stderr,"\nwriting <%c> but first sleep %f seconds\n",*sp,t);*/
 		/* skip sleep before writing first character */
 		if (sp != buffer) exp_usleep((long)(t*1000000));
 
@@ -854,6 +1048,289 @@ struct human_arg *arg;
 	return(0);
 }
 
+struct exp_i *exp_i_pool = 0;
+struct exp_fd_list *exp_fd_list_pool = 0;
+
+#define EXP_I_INIT_COUNT	10
+#define EXP_FD_INIT_COUNT	10
+
+struct exp_i *
+exp_new_i(interp)
+Tcl_Interp *interp;
+{
+	int n;
+	struct exp_i *i;
+
+	if (!exp_i_pool) {
+		/* none avail, generate some new ones */
+		exp_i_pool = i = (struct exp_i *)ckalloc(
+			EXP_I_INIT_COUNT * sizeof(struct exp_i));
+		for (n=0;n<EXP_I_INIT_COUNT-1;n++,i++) {
+			i->next = i+1;
+		}
+		i->next = 0;
+	}
+
+	/* now that we've made some, unlink one and give to user */
+
+	i = exp_i_pool;
+	exp_i_pool = exp_i_pool->next;
+	return i;
+}
+
+struct exp_fd_list *
+exp_new_fd(interp,val)
+Tcl_Interp *interp;
+int val;
+{
+	int n;
+	struct exp_fd_list *fd;
+
+	if (!exp_fd_list_pool) {
+		/* none avail, generate some new ones */
+		exp_fd_list_pool = fd = (struct exp_fd_list *)ckalloc(
+			EXP_FD_INIT_COUNT * sizeof(struct exp_fd_list));
+		for (n=0;n<EXP_FD_INIT_COUNT-1;n++,fd++) {
+			fd->next = fd+1;
+		}
+		fd->next = 0;
+	}
+
+	/* now that we've made some, unlink one and give to user */
+
+	fd = exp_fd_list_pool;
+	exp_fd_list_pool = exp_fd_list_pool->next;
+	fd->fd = val;
+	return fd;
+}
+
+void
+exp_free_fd(fd_first)
+struct exp_fd_list *fd_first;
+{
+	struct exp_fd_list *fd, *penultimate;
+
+	if (!fd_first) return;
+
+	/* link entire chain back in at once by first finding last pointer */
+	/* making that point back to pool, and then resetting pool to this */
+
+	/* run to end */
+	for (fd = fd_first;fd;fd=fd->next) {
+		penultimate = fd;
+	}
+	penultimate->next = exp_fd_list_pool;
+	exp_fd_list_pool = fd_first;
+}
+
+void
+exp_free_i(i)
+struct exp_i *i;
+{
+	if (i->next) exp_free_i(i->next);
+
+	exp_free_fd(i->fd_list);
+
+	/* here's the long form */
+	/* if dynamic & direct  free(var)  free(val) */
+	/*	1	  1			1	*/
+	/*	1	  0	    1		1	*/
+	/*	0	  1				*/
+	/*	0	  0			1	*/
+
+	/* if i->variable was a bogus variable name, i->value might not be */
+	/* set, so test i->value to protect this */
+
+	if ((i->dynamic == i->direct) && i->value) free(i->value);
+	else if (i->dynamic) {
+		if (i->value) free(i->value);
+		free(i->variable);
+	}
+
+	i->next = exp_i_pool;
+	exp_i_pool = i;
+}
+
+
+/* generate a descriptor for a "-i" flag */
+struct exp_i *
+exp_new_i_complex(interp,/*direct,*/arg,dynamic)
+Tcl_Interp *interp;
+/*int direct;*/
+char *arg;
+int dynamic;		/* if we have to copy the args */
+			/* should only need do this in expect_before/after */
+{
+	struct exp_i *i;
+	char **stringp;
+
+	i = exp_new_i(interp);
+
+/*	i->direct = direct;*/
+	i->direct = !!isdigit(arg[0]);
+	if (i->direct) {
+		stringp = &i->value;
+	} else {
+		i->value = 0;
+		stringp = &i->variable;
+	}
+
+	i->dynamic = dynamic;
+	if (dynamic) {
+		*stringp = ckalloc(strlen(arg)+1);
+		strcpy(*stringp,arg);
+	} else {
+		*stringp = arg;
+	}
+
+	i->next = 0;
+	i->fd_list = 0;
+	if (TCL_ERROR == exp_i_update(interp,i)) {
+		exp_free_i(i);
+		return 0;
+	}
+
+	return i;
+}
+
+void
+exp_i_add_fd(interp,i,fd)
+Tcl_Interp *interp;
+struct exp_i *i;
+int fd;
+{
+	struct exp_fd_list *new_fd;
+
+	new_fd = exp_new_fd(interp,fd);
+	new_fd->next = i->fd_list;
+	i->fd_list = new_fd;
+}
+
+/* this routine assumes i->fd is meaningful */
+void
+exp_i_parse_fds(interp,i)
+Tcl_Interp *interp;
+struct exp_i *i;
+{
+	char *p = i->value;
+
+	/* reparse it */
+	while (1) {
+		int m;
+		int negative = 0;
+		int valid_spawn_id = 0;
+
+		m = 0;
+		while (isspace(*p)) p++;
+		for (;;p++) {
+			if (*p == '-') negative = 1;
+			else if (isdigit(*p)) {
+				m = m*10 + (*p-'0');
+				valid_spawn_id = 1;
+			} else if (*p == '\0' || isspace(*p)) break;
+		}
+
+		/* we either have a spawn_id or whitespace at end of string */
+
+		/* skip whitespace end-of-string */
+		if (!valid_spawn_id) break;
+
+		if (negative) m = -m;
+
+		exp_i_add_fd(interp,i,m);
+	}
+}
+	
+/* updates a single exp_i struct */
+/* returns TCL_OK or TCL_ERROR */
+int
+exp_i_update(interp,i)
+Tcl_Interp *interp;
+struct exp_i *i;
+{
+	char *p;	/* string representation of list of spawn ids */
+
+	if (!i->direct) {
+		p = exp_get_var(interp,i->variable);
+		if (!p) {
+			exp_error(interp,"%s undefined",i->variable);
+			return TCL_ERROR;
+		}
+
+		if (i->value) {
+			if (streq(p,i->value)) return TCL_OK;
+
+			/* replace new value with old */
+			free(i->value);
+		}
+		i->value = ckalloc(strlen(p)+1);
+		strcpy(i->value,p);
+
+		exp_free_fd(i->fd_list);
+		i->fd_list = 0;
+	} else {
+		/* no free, because this should only be called on */
+		/* "direct" i's once */
+		i->fd_list = 0;
+	}
+	exp_i_parse_fds(interp,i);
+
+	return TCL_OK;
+}
+
+struct exp_i *
+exp_new_i_simple(interp,fd)
+Tcl_Interp *interp;
+int fd;
+{
+	struct exp_i *i;
+
+	i = exp_new_i(interp);
+	i->next = 0;
+	i->fd_list = 0;
+
+	i->direct = TRUE;
+	i->dynamic = FALSE;
+
+	exp_i_add_fd(interp,i,fd);
+
+	return i;
+}
+
+/*ARGSUSED*/
+static int
+cmdSendLog(clientData, interp, argc, argv)
+ClientData clientData;
+Tcl_Interp *interp;
+int argc;
+char **argv;
+{
+	char *string;
+	int len;
+
+	argv++;
+	argc--;
+	while (argc) {
+		if (streq(*argv,"--")) {
+			argc--; argv++;
+			break;
+		} else break;
+	}
+	if (argc != 1) {
+		exp_error(interp,"usage: send [args] string");
+		return TCL_ERROR;
+	}
+
+	string = *argv;
+
+	len = strlen(string);
+
+	if (debugfile) fwrite(string,1,len,debugfile);
+	if (logfile) fwrite(string,1,len,logfile);
+
+	return(TCL_OK);
+}
+
 
 /* I've rewritten this to be unbuffered.  I did this so you could shove */
 /* large files through "send".  If you are concerned about efficiency */
@@ -866,31 +1343,38 @@ Tcl_Interp *interp;
 int argc;
 char **argv;
 {
-	int master = -1;
-	int master_flag = FALSE;
-	int i;
+	int m = -1;	/* spawn id (master) */
 	int wc = 0;	/* if negative, a write has failed */
-	extern int is_debugging;
+	int rc; 	/* final result of this procedure */
 	struct human_arg human_args;
 	struct slow_arg slow_args;
 #define SEND_STYLE_PLAIN	0
 #define SEND_STYLE_HUMAN	1
 #define SEND_STYLE_SLOW		2
-#define SEND_STYLE_NONE		3
+#define SEND_STYLE_ZERO		3
 	int send_style = SEND_STYLE_PLAIN;
 	int want_cooked = TRUE;
+	char *string;		/* string to send */
+	int len;		/* length of string to send */
+	int zeros;		/* count of how many ascii zeros to send */
+
+	char *i_masters = 0;
+	struct exp_fd_list *fd;
+	struct exp_i *i;
 
 	argv++;
 	argc--;
 	while (argc) {
-		if (streq(*argv,"-i")) {
+		if (streq(*argv,"--")) {
+			argc--; argv++;
+			break;
+		} else if (streq(*argv,"-i")) {
 			argc--; argv++;
 			if (argc==0) {
-				exp_error(interp,"-i requires following spawn_id");
+				exp_error(interp,"usage: -i spawn_id");
 				return(TCL_ERROR);
 			}
-			master = atoi(*argv);
-			master_flag = TRUE;
+			i_masters = *argv;
 			argc--; argv++;
 			continue;
 		} else if (streq(*argv,"-h")) {
@@ -903,118 +1387,122 @@ char **argv;
 			if (-1 == get_slow_args(interp,&slow_args))
 				return(TCL_ERROR);
 			send_style = SEND_STYLE_SLOW;
+		} else if (streq(*argv,"-0")) {
+			argc--; argv++;
+			if (!*argv) {
+				exp_error(interp,"usage: -0 integer");
+				return TCL_ERROR;
+			}
+			send_style = SEND_STYLE_ZERO;
+			zeros = atoi(*argv);
+			if (zeros < 1) return TCL_OK;
+			argc--; argv++;
 		} else if (streq(*argv,"-raw")) {
 			argc--; argv++;
 			want_cooked = FALSE;
 		} else break;
 	}
+	if (argc != 1) {
+		exp_error(interp,"usage: send [args] string");
+		return TCL_ERROR;
+	}
 
-	if (clientData == &sendCD_user) master = 1;
-	else if (clientData == &sendCD_error) master = 2;
-	else if (clientData == &sendCD_tty) master = dev_tty;
-	else if ((clientData != &sendCD_log) && !master_flag) {
+	if (send_style == SEND_STYLE_ZERO) {
+		string = "<zero(s)>";
+	} else {
+		string = *argv;
+	}
+	len = strlen(string);
+
+	if (clientData == &sendCD_user) m = 1;
+	else if (clientData == &sendCD_error) m = 2;
+	else if (clientData == &sendCD_tty) m = exp_dev_tty;
+	else if (!i_masters) {
 		/* we really do want to check if it is open */
 		/* but since stdin could be closed, we have to first */
 		/* get the fd and then convert it from 0 to 1 if necessary */
-		if (0 == exp_update_master(interp,&master,0,0))
+		if (0 == exp_update_master(interp,&m,0,0))
 			return(TCL_ERROR);
-		/* true if called as Send with user_spawn_id */
-		if (is_user(master)) {
-			clientData = &sendCD_user;
-			master = 1;
 		}
-	}
-	/* now do the close check */
-	if (fs[master].flags & FD_CLOSED) {
-		exp_error(interp,"%d: invalid spawn id",master);
-		return(TCL_ERROR);
+
+	/* if master != -1, then it holds desired master */
+	/* else i_masters does */
+
+	if (m != -1) {
+		i = exp_new_i_simple(interp,m);
+	} else {
+		i = exp_new_i_complex(interp /*,1*/ /*direct*/,i_masters,FALSE);
+		if (i == 0) return TCL_ERROR;
 	}
 
-/*#define send_to_stdout	(clientData == &sendCD_user)*/
 #define send_to_stderr	(clientData == &sendCD_error)
 #define send_to_proc	(clientData == &sendCD_proc)
-#define send_to_log	(clientData == &sendCD_log)
 #define send_to_user	((clientData == &sendCD_user) || \
 			 (clientData == &sendCD_tty))
 
-/*debuglog("master = %d  send_to_proc = %d\r\n",master,send_to_proc);*/
-
-	if (send_to_log) send_style = SEND_STYLE_NONE;
-	if (send_to_proc) want_cooked = FALSE;
-	if (send_to_proc) debuglog("send: sent {");
-	/* if the closing brace doesn't appear, that's because an error was */
-	/* encountered before we could send it */
-
-	for (i = 0;i<argc;i++) {
-		if (i != 0) {
-			/* wedge ' ' between other args.  If you don't want */
-			/* this, make it all into one arg! */
 			if (send_to_proc) {
-				debuglog(" ");
+		want_cooked = FALSE;
+		debuglog("send: sending {%s} to {",dprintify(string));
+		/* if closing brace doesn't appear, that's because an error */
+		/* was encountered before we could send it */
 			} else {
-				if (debugfile) fwrite(" ",1,1,debugfile);
+		if (debugfile)
+			fwrite(string,1,len,debugfile);
 				if ((send_to_user && logfile_all) || logfile)
-					fwrite(" ",1,1,logfile);
+			fwrite(string,1,len,logfile);
 			}
-			switch (send_style) {
-			case SEND_STYLE_SLOW:
-				wc = slow_write(master," ",1,&slow_args);
-				break;
-			case SEND_STYLE_HUMAN:
-				wc = human_write(master," ",&human_args);
-				break;
-			case SEND_STYLE_NONE:
-				break;
-			default:
-				wc = exact_write(master," ",1);
-				break;
+
+	for (fd=i->fd_list;fd;fd=fd->next) {
+		m = fd->fd;
+
+		if (send_to_proc) {
+			debuglog(" %d ",m);
 			}
+
+		/* true if called as Send with user_spawn_id */
+		if (exp_is_stdinfd(m)) m = 1;
+
+		/* check validity of each - i.e., are they open */
+		if (0 == exp_fd2f(interp,m,1,0,"send")) {
+			rc = TCL_ERROR;
+			goto finish;
 		}
-		if (wc >= 0) {
-			int len = strlen(argv[i]);
-			if (send_to_proc) {
-				debuglog("%s",dprintify(argv[i]));
-			} else {
-				if (debugfile)
-					fwrite(argv[i],1,len,debugfile);
-				if ((send_to_user && logfile_all) || logfile)
-					fwrite(argv[i],1,len,logfile);
+		/* Check if Tcl is using a different fd for output */
+		if (fs[m].tcl_handle) {
+			m = fs[m].tcl_output;
 			}
 
-			if (want_cooked) argv[i] = exp_cook(argv[i],&len);
+		if (want_cooked) string = exp_cook(string,&len);
 
 			switch (send_style) {
 			case SEND_STYLE_SLOW:
-				wc = slow_write(master,argv[i],len,&slow_args);
+			wc = slow_write(m,string,len,&slow_args);
 				break;
 			case SEND_STYLE_HUMAN:
-				wc = human_write(master,argv[i],&human_args);
+			wc = human_write(m,string,&human_args);
 				break;
-			case SEND_STYLE_NONE:
+		case SEND_STYLE_ZERO:
+			for (;zeros>0;zeros--) wc = write(m,"",1);
+			/* catching error on last write is sufficient */
 				break;
 			default:
-				wc = exact_write(master,argv[i],len);
+			wc = exact_write(m,string,len);
 				break;
 			}
-		}
+
 		if (wc < 0) {
-
-/* actually, I think close is the wrong thing to do here */
-#if 0
-			int save_errno = errno;
-
-			/* write failed, so close fd.  But if close fails */
-			/* describe write error rather than close */
-			if (send_to_proc) exp_close(interp,master);
-#endif
 			exp_error(interp,"write(spawn_id=%d): %s",
-					master,sys_errlist[errno]);
-			return(TCL_ERROR);
+					m,sys_errlist[errno]);
+			rc = TCL_ERROR;
+			goto finish;
 		}
 	}
-	if (send_to_proc) debuglog("} to spawn id %d\r\n",master);
+	if (send_to_proc) debuglog("}\r\n");
 
-	return(TCL_OK);
+	rc = TCL_OK;
+ finish:
+	exp_free_i(i);
+	return rc;
 }
 
 void
@@ -1079,6 +1567,7 @@ char **argv;
 			logfile_all = FALSE;
 			return(TCL_ERROR);
 		}
+		exp_close_on_exec(logfile);
 /*new*/		setbuf(logfile,(char *)0);
 	}
 	return(TCL_OK);
@@ -1092,13 +1581,15 @@ Tcl_Interp *interp;
 int argc;
 char **argv;
 {
-	if (argc != 2) {
-		exp_error(interp,"usage: log_user expr");
-		return(TCL_ERROR);
-	}
+	int old_loguser = loguser;
 
+	if (argc == 2) {
 	if (0 == atoi(argv[1])) loguser = FALSE;
 	else loguser = TRUE;
+	}
+
+	sprintf(interp->result,"%d",old_loguser);
+
 	return(TCL_OK);
 }
 
@@ -1137,6 +1628,7 @@ char **argv;
 			exp_error(interp,"%s: %s",*argv,sys_errlist[errno]);
 			return(TCL_ERROR);
 		}
+		exp_close_on_exec(debugfile);
 		setbuf(debugfile,(char *)0);
 		fopened = TRUE;
 		argc--;argv++;
@@ -1190,28 +1682,57 @@ Tcl_Interp *interp;
 int argc;
 char **argv;
 {
+	int onexec_flag = FALSE;	/* true if -onexec seen */
+	int close_onexec;
 	int m = -1;
 
-	if (argc == 1) {
-		if (exp_update_master(interp,&m,1,0) == 0) return(TCL_ERROR);
-	} else if (streq(argv[1],"-i")) {
-		if (argc != 3) {
-			exp_error(interp,"usage: close [-i spawn_id]");
+	int argc_orig = argc;
+	char **argv_orig = argv;
+
+	argc--; argv++;
+
+	for (;argc>0;argc--,argv++) {
+		if (streq(*argv,"-i")) {
+			argc--; argv++;
+			if (*argv) {
+				exp_error(interp,"usage: -i spawn_id");
+				return(TCL_ERROR);
+			}
+			m = atoi(*argv);
+		} else if (streq(*argv,"-onexec")) {
+			argc--; argv++;
+			if (*argv) {
+				exp_error(interp,"usage: -onexec 0|1");
 			return(TCL_ERROR);
 		}
-		m = atoi(argv[2]);
+			onexec_flag = TRUE;
+			close_onexec = atoi(*argv);
+		} else break;
 	}
 
-	if (m != -1) return(exp_close(interp,m));
+	if (argc) {
+		/* doesn't look like our format, it must be a Tcl-style file */
+		/* handle.  Lucky that formats are easily distinguishable. */
+		/* Historical note: we used "close"  long before there was a */
+		/* Tcl builtin by the same name. */
 
-	/* doesn't look like our format, it must be a Tcl-style file handle */
-	/* we're so lucky that the formats are easily distinguishable */
-	/* Historical note: we used "close" in Tcl long before there was a */
-	/* builtin by the same name. */
+		/* don't know what is formally correct as 1st arg, but I see */
+		/* the code doesn't use it anyway */
+		return(Tcl_CloseCmd(clientData,interp,argc_orig,argv_orig));
+	}
 
-	/* don't know what is formally correct as 1st arg, but I see the */
-	/* code doesn't use it anyway */
-	return(Tcl_CloseCmd(clientData,interp,argc,argv));
+	if (m == -1) {
+		if (exp_update_master(interp,&m,1,0) == 0) return(TCL_ERROR);
+	}
+
+	if (onexec_flag) {
+		/* heck, don't even bother to check if fd is open or a real */
+		/* spawn id, nothing else depends on it */
+		fcntl(m,F_SETFD,close_onexec);
+		return TCL_OK;
+	}
+
+	return(exp_close(interp,m));
 }
 
 /*ARGSUSED*/
@@ -1302,7 +1823,10 @@ char **argv;
 
 	if (!(f = exp_fd2f(interp,master,0,0,"wait"))) return(TCL_ERROR);
 
-	if (!(f->flags & FD_SYSWAITED)) {
+	/* hmm, user called wait without calling close? */
+	if (!f->user_closed) exp_close(interp,master);
+
+	if (!(f->sys_waited)) {
 #if defined(NOWAITPID) && defined(HAVE_WAIT4)
 # undef NOWAITPID
 # define waitpid(pid,stat,opt) wait4(pid,stat,opt,NULL)
@@ -1323,17 +1847,18 @@ char **argv;
 				debuglog("wait found unknown pid %d\r\n",pid);
 				continue;	/* drop on floor */
 			}
-			fs[i].flags |= FD_SYSWAITED;
+			fs[i].sys_waited = TRUE;
 			fs[i].wait = status;
 		}
 		f->wait = status;
 #else
 		pid = waitpid(f->pid,&f->wait,0);
 #endif
-		f->flags |= FD_SYSWAITED;
+		f->sys_waited = TRUE;
 	}
-	if (f->flags & FD_CLOSED) f->flags = 0;
-	f->flags |= FD_USERWAITED;
+
+	exp_destroy(master);
+
 #ifndef WEXITSTATUS
 #define WEXITSTATUS(x) x
 #endif
@@ -1374,6 +1899,8 @@ Tcl_Interp *interp;
 int argc;
 char **argv;
 {
+	/* tell Saber to ignore non-use of ttyfd */
+	/*SUPPRESS 591*/
 	int ttyfd;
 
 	if (argc > 1) {
@@ -1395,14 +1922,14 @@ char **argv;
 	/* and other places */
 	signal(SIGHUP,SIG_IGN);
 
-	/* freopen(,,stdin) prevents confusion between send/expect_user */
+	/* reopen prevents confusion between send/expect_user */
 	/* accidentally mapping to a real spawned process after a disconnect */
-	/* A better solution might be to put code in the _user routines to */
-	/* check that a disconnect has occured. */
-	/* freopen(,,stderr) saves error checking in error/log routines. */
-	freopen("/dev/null","r",stdin);
-	freopen("/dev/null","w",stdout);
-	freopen("/dev/null","w",stderr);
+	exp_close(interp,0);	open("/dev/null",0);	fd_new(0, EXP_NOPID);
+	exp_close(interp,1);	open("/dev/null",1);	fd_new(1, EXP_NOPID);
+	/* reopen stderr saves error checking in error/log routines. */
+	exp_close(interp,2);	open("/dev/null",1);	fd_new(2, EXP_NOPID);
+
+	Tcl_UnsetVar(interp,"tty_spawn_id",TCL_GLOBAL_ONLY);
 
 #ifdef DO_SETSID
 	setsid();
@@ -1478,8 +2005,7 @@ char **argv;
 	}
 	command = argv[0];
 	if (dash_name) {
-		argv[0] = malloc(1+strlen(command));
-		/* check for malloc failure */
+		argv[0] = ckalloc(1+strlen(command));
 		sprintf(argv[0],"-%s",command);
 	}
 
@@ -1510,14 +2036,8 @@ char **argv;
 		return(TCL_ERROR);
 	}
 
-	if (0 == (masters = (int *)malloc((argc-1)*sizeof(int)))) {
-		exp_error(interp,"malloc(%d spawn_id's)",(argc-1));
-		return(TCL_ERROR);
-	}
-	if (0 == (masters2 = (int *)malloc((argc-1)*sizeof(int)))) {
-		exp_error(interp,"malloc(%d spawn_id's)",(argc-1));
-		return(TCL_ERROR);
-	}
+	masters = (int *)ckalloc((argc-1)*sizeof(int));
+	masters2 = (int *)ckalloc((argc-1)*sizeof(int));
 
 	for (i=1;i<argc;i++) {
 		j = atoi(argv[i]);
@@ -1594,16 +2114,22 @@ char **argv;
 	int rc = TCL_RETURN;
 
 	if (argc == 1) return(TCL_RETURN);
-	if (argc > 3) {
-		exp_error(interp,"usage: return [-tcl] [value]\n");
-		return(TCL_ERROR);
+
+	argc--; argv++;
+
+	if (argc && streq(*argv,"-tcl")) {
+		rc = TCL_RETURN_TCL;
+		argc--; argv++;
 	}
 
-	if (streq(argv[1],"-tcl")) rc = TCL_RETURN_TCL;
+	if (argc) {
+		Tcl_SetResult(interp,*argv,TCL_VOLATILE);
+		argc--; argv++; 
+	}
 
-	if (!(rc == TCL_RETURN_TCL && argc == 2)) {
-		/* if either return "-tcl expr" or "expr", get the expr */
-		Tcl_SetResult(interp,argv[1],TCL_VOLATILE);
+	if (argc) {
+		exp_error(interp,"usage: too many args\n");
+		return(TCL_ERROR);
 	}
 
 	return(rc);
@@ -1622,12 +2148,21 @@ Tcl_Interp *interp;
 
 	Tcl_CreateCommand(interp,"close",
 		cmdClose,(ClientData)0,exp_deleteProc);
+	Tcl_CreateCommand(interp,"exp_close",
+		cmdClose,(ClientData)0,exp_deleteProc);
 	Tcl_CreateCommand(interp,"debug",
 		cmdDebug,(ClientData)0,exp_deleteProc);
+	Tcl_CreateCommand(interp,"exp_debug",
+		cmdDebug,(ClientData)0,exp_deleteProc);
 	Tcl_CreateCommand(interp,"disconnect",
 		cmdDisconnect,(ClientData)0,exp_deleteProc);
+	Tcl_CreateCommand(interp,"exp_disconnect",
+		cmdDisconnect,(ClientData)0,exp_deleteProc);
 	Tcl_CreateCommand(interp,"exit",
 		cmdExit,(ClientData)0,exp_deleteProc);
+	Tcl_CreateCommand(interp,"exp_exit",
+		cmdExit,(ClientData)0,exp_deleteProc);
+
 
 	/* following commands are created in exp_init_expect in expect.c */
 	/* to prevent unnecessary name space pollution by the ClientData */
@@ -1638,47 +2173,85 @@ Tcl_Interp *interp;
 
 	Tcl_CreateCommand(interp,"continue",
 		cmdContinueExpect,(ClientData)0,exp_deleteProc);
+	Tcl_CreateCommand(interp,"exp_continue",
+		cmdContinueExpect,(ClientData)0,exp_deleteProc);
 	Tcl_CreateCommand(interp,"expect_version",
 		cmdExpectVersion,(ClientData)0,exp_deleteProc);
 	Tcl_CreateCommand(interp,"fork",
 		cmdFork,(ClientData)0,exp_deleteProc);
+	Tcl_CreateCommand(interp,"exp_fork",
+		cmdFork,(ClientData)0,exp_deleteProc);
+	Tcl_CreateCommand(interp,"pid",
+		cmdGetpid,(ClientData)0,exp_deleteProc);
+	Tcl_CreateCommand(interp,"exp_pid",
+		cmdExpPid,(ClientData)0,exp_deleteProc);
 	Tcl_CreateCommand(interp,"getpid",
 		cmdGetpid,(ClientData)0,exp_deleteProc);
 	Tcl_CreateCommand(interp,"interact",
 		cmdInteract,(ClientData)0,exp_deleteProc);
 	Tcl_CreateCommand(interp,"interpreter",
 		cmdInterpreter,(ClientData)0,exp_deleteProc);
+	Tcl_CreateCommand(interp,"exp_interpreter",
+		cmdInterpreter,(ClientData)0,exp_deleteProc);
 	Tcl_CreateCommand(interp,"log_file",
 		cmdLogFile,(ClientData)0,exp_deleteProc);
+	Tcl_CreateCommand(interp,"exp_log_file",
+		cmdLogFile,(ClientData)0,exp_deleteProc);
 	Tcl_CreateCommand(interp,"log_user",
 		cmdLogUser,(ClientData)0,exp_deleteProc);
+	Tcl_CreateCommand(interp,"exp_log_user",
+		cmdLogUser,(ClientData)0,exp_deleteProc);
 	Tcl_CreateCommand(interp,"overlay",
 		cmdOverlay,(ClientData)0,exp_deleteProc);
+	Tcl_CreateCommand(interp,"exp_overlay",
+		cmdOverlay,(ClientData)0,exp_deleteProc);
 	Tcl_CreateCommand(interp,"prompt1",
 		cmdPrompt1,(ClientData)0,exp_deleteProc);
+	Tcl_CreateCommand(interp,"exp_prompt1",
+		cmdPrompt1,(ClientData)0,exp_deleteProc);
 	Tcl_CreateCommand(interp,"prompt2",
 		cmdPrompt2,(ClientData)0,exp_deleteProc);
+	Tcl_CreateCommand(interp,"exp_prompt2",
+		cmdPrompt2,(ClientData)0,exp_deleteProc);
 	Tcl_CreateCommand(interp,"return",
 		cmdReturnInter,(ClientData)0,exp_deleteProc);
+	Tcl_CreateCommand(interp,"exp_return",
+		cmdReturnInter,(ClientData)0,exp_deleteProc);
 	Tcl_CreateCommand(interp,"send",
 		cmdSend,(ClientData)&sendCD_proc,exp_deleteProc);
+	Tcl_CreateCommand(interp,"exp_send",
+		cmdSend,(ClientData)&sendCD_proc,exp_deleteProc);
 	/* Tk may wipe out "send" so provide an alias: "send_spawn" */
-	Tcl_CreateCommand(interp,"send_spawn",
+	Tcl_CreateCommand(interp,"exp_send_spawn",
 		cmdSend,(ClientData)&sendCD_proc,exp_deleteProc);
 	Tcl_CreateCommand(interp,"send_error",
 		cmdSend,(ClientData)&sendCD_error,exp_deleteProc);
+	Tcl_CreateCommand(interp,"exp_send_error",
+		cmdSend,(ClientData)&sendCD_error,exp_deleteProc);
 	Tcl_CreateCommand(interp,"send_log",
-		cmdSend,(ClientData)&sendCD_log,exp_deleteProc);
+		cmdSendLog,(ClientData)0,exp_deleteProc);
+	Tcl_CreateCommand(interp,"exp_send_log",
+		cmdSendLog,(ClientData)0,exp_deleteProc);
 	Tcl_CreateCommand(interp,"send_tty",
 		cmdSend,(ClientData)&sendCD_tty,exp_deleteProc);
+	Tcl_CreateCommand(interp,"exp_send_tty",
+		cmdSend,(ClientData)&sendCD_tty,exp_deleteProc);
 	Tcl_CreateCommand(interp,"send_user",
 		cmdSend,(ClientData)&sendCD_user,exp_deleteProc);
+	Tcl_CreateCommand(interp,"exp_send_user",
+		cmdSend,(ClientData)&sendCD_user,exp_deleteProc);
 	Tcl_CreateCommand(interp,"spawn",
 		cmdSpawn,(ClientData)0,exp_deleteProc);
 	Tcl_CreateCommand(interp,"strace",
 		cmdTrace,(ClientData)0,exp_deleteProc);
+	Tcl_CreateCommand(interp,"exp_strace",
+		cmdTrace,(ClientData)0,exp_deleteProc);
 	Tcl_CreateCommand(interp,"system",
 		cmdSystem,(ClientData)0,exp_deleteProc);
+	Tcl_CreateCommand(interp,"exp_system",
+		cmdSystem,(ClientData)0,exp_deleteProc);
 	Tcl_CreateCommand(interp,"wait",
 		cmdWait,(ClientData)0,exp_deleteProc);
+	Tcl_CreateCommand(interp,"exp_wait",
+		cmdWait,(ClientData)0,exp_deleteProc);
 }
