@@ -30,6 +30,7 @@
 
  */
 
+#include "tcl.h"
 #include "exp_conf.h"
 #include "regexp.h"
 #include "exp_global.h"
@@ -1007,6 +1008,7 @@ char *p;
 	register int count = 0;
 	register char *scan;
 	register char *opnd;
+/*DEL*/	extern char *strchr();
 
 	scan = reginput;
 	opnd = OPERAND(p);
