@@ -6,8 +6,8 @@ Design and implementation of this program was paid for by U.S. tax
 dollars.  Therefore it is public domain.  However, the author and NIST
 would appreciate credit if this program or parts of it are used.
 
-$Revision: 1.12 $
-$Date: 1993/04/26 22:54:50 $
+$Revision: 1.17 $
+$Date: 1993/08/20 04:29:28 $
 
 */
 
@@ -19,9 +19,13 @@ $Date: 1993/04/26 22:54:50 $
 #include <ctype.h>	/* for isspace */
 #include <time.h>	/* for time(3) */
 #include <setjmp.h>
-#include <sys/wait.h>
+
 #include "exp_conf.h"
 
+#ifdef HAVE_SYS_WAIT_H
+#include <sys/wait.h>
+#endif
+
 #include "tcl.h"
 extern char *tclRegexpError;	/* declared in tclInt.h */
 
@@ -35,6 +39,9 @@ extern char *tclRegexpError;	/* declared in tclInt.h */
 #include "exp_main.h"
 #include "exp_event.h"
 #include "exp_tty.h"
+#ifdef TCL_DEBUGGER
+#include "Dbg.h"
+#endif
 
 /* initial length of strings that we can guarantee patterns can match */
 int exp_default_match_max =	2000;
@@ -84,21 +91,19 @@ static RETSIGTYPE
 sigint_handler(n)
 int n;			/* unused, for compatibility with STDC */
 {
-#ifdef EXP_DEBUGGER
-	extern int exp_debugger_active;
-#endif
+	extern Tcl_Interp *exp_interp;
 
 #ifdef REARM_SIG
 	signal(SIGINT,sigint_handler);/* not nec. for BSD, but doesn't hurt */
 #endif
 
-#ifdef EXP_DEBUGGER
-	if (exp_debugger_active) {
+#ifdef TCL_DEBUGGER
+	if (exp_tcl_debugger_available) {
 		/* if the debugger is active and we're reading something, */
 		/* force the debugger to go interactive now and when done, */
-		/* restart the read.  
+		/* restart the read.  */
 
-		debugger(interp,env_valid);
+		Dbg_On(exp_interp,env_valid);
 
 		/* restart the read */
 		if (env_valid) longjmp(env,2);
@@ -124,10 +129,8 @@ int n;			/* unused, for compatibility with STDC */
 eof/timeout don't use any of theirs, but the algorithm is simpler this way. */
 
 struct ecase {	/* case for expect command */
-	int m;		/* master */
-	int n;		/* number of patterns */
+	struct exp_i	*i_list;
 	char *pat;	/* original pattern spec */
-	char **patn;	/* if multiple patterns in list */
 	char *body;	/* ptr to body to be executed upon match */
 #define PAT_EOF		1
 #define PAT_TIMEOUT	2
@@ -138,31 +141,28 @@ struct ecase {	/* case for expect command */
 	int use;	/* PAT_XXX */
 	int glob_start;	/* start of string matched when use == PAT_GLOB */
 	int transfer;	/* if false, leave matched chars in input stream */
+	int iwrite;	/* if true write spawn_id */
+	int iread;	/* if true, reread indirects */
+	int timestamp;	/* if true, write timestamps */
 #define CASE_UNKNOWN	0
 #define CASE_NORM	1
 #define CASE_LOWER	2
 	int Case;	/* convert case before doing match? */
 	regexp *re;	/* if this is 0, then pattern match via glob */
-#define EXP_BEFORE	1
-#define EXP_DURING	2
-#define EXP_AFTER	3
-	int owner;	/* one of the above */
 };
 
 /* data structure for saving results of expect_before/after */
-struct expect_special {
+struct expect_global {
 	char *prefix;		/* command plus blank to shove in front */
 				/* of args upon finding 1 arg and recursing */
 	struct ecase *ecases;
 	int ecount;		/* count of cases */
-	int *masters;
-	int mcount;		/* number of masters */
-	int me;
+	struct exp_i *i_list;
 };
 
-static struct expect_special
-	before = {"expect_before ", NULL, 0, NULL, 0, EXP_BEFORE},
-	after  = {"expect_after ",  NULL, 0, NULL, 0, EXP_AFTER};
+static struct expect_global
+	before = {"expect_before ", NULL, 0, NULL},
+	after  = {"expect_after ",  NULL, 0, NULL};
 
 /* remove nulls from s.  Initially, the number of chars in s is c, */
 /* not strlen(s).  This count does not include the trailing null. */
@@ -188,66 +188,19 @@ int c;
 	return(count);
 }
 
-/* generate printable versions of random ASCII strings.  This is used by */
-/* cmdExpect when -d forces it to print strings it is examining. */
-char *
-printify(s)
-char *s;
-{
-	static int destlen = 0;
-	static char *dest = 0;
-	char *d;		/* ptr into dest */
-	unsigned int need;
-
-	if (s == 0) return("<null>");
-
-	/* worst case is every character takes 3 to printify */
-	need = strlen(s)*3 + 1;
-	if (need > destlen) {
-		if (dest) free(dest);
-		if (!(dest = malloc(need))) {
-			destlen = 0;
-			return("malloc failed in printify");
-		}
-		destlen = need;
-	}
-
-	for (d = dest;*s;s++) {
-		if (*s == '\r') {
-			strcpy(d,"\\r");		d += 2;
-		} else if (*s == '\n') {
-			strcpy(d,"\\n");		d += 2;
-		} else if (*s == '\t') {
-			strcpy(d,"\\t");		d += 2;
-		} else if ((unsigned)*s < 0x20) { /* unsigned strips parity */
-			sprintf(d,"\\C%c",*s + '`');	d += 3;
-		} else if (*s == 0x7f) {
-			/* not syntactically correct, but you get the point */
-			strcpy(d,"\\7f");		d += 3;
-		} else {
-			*d = *s;			d += 1;
-		}
-	}
-	*d = '\0';
-	return(dest);
-}
-
 /* free up any argv structures in the ecases */
 static void
-free_ecases(ecases,ecases_inuse,me)
+free_ecases(ecases,ecases_inuse,before_or_after)
 struct ecase *ecases;
 int ecases_inuse;
-int me;
+int before_or_after;
 {
 	int i;
 	struct ecase *ec;
 
 	for (ec=ecases,i=0;i<ecases_inuse;i++,ec++) {
 		if (ec->re) free((char *)ec->re);
-		/* individual elements of each list don't have to be freed */
-		/* because SplitList allocates them all from single blocks! */
-		if (ec->patn) free((char *)ec->patn);
-		if (me != EXP_DURING) {
+		if (before_or_after) {
 			if (ec->pat) free(ec->pat);
 			if (ec->body) free(ec->body);
 		}
@@ -261,52 +214,27 @@ int me;
 static char *exp_strdup(s)
 char *s;
 {
-	char *news = malloc(strlen(s) + 1);
-	if (news) strcpy(news,s);
+	char *news = ckalloc(strlen(s) + 1);
+	strcpy(news,s);
 	return(news);
 }
 #endif
 
 /* In many places, there is no need to malloc a copy of a string, since it */
 /* will be freed before we return to Tcl */
-#define SUCCESS 0
-#define FAILURE 1
-static int
-save_str(lhs,rhs,me)
+static void
+save_str(lhs,rhs,nosave)
 char **lhs;	/* left hand side */
 char *rhs;	/* right hand side */
-int me;
+int nosave;
 {
-	if ((me == EXP_DURING) || (rhs == 0)) {
+	if (nosave || (rhs == 0)) {
 		*lhs = rhs;
-		return SUCCESS;
-	}
-	*lhs = malloc(strlen(rhs) + 1);
-	if (!*lhs) return FAILURE;
+	} else {
+		*lhs = ckalloc(strlen(rhs) + 1);
 	strcpy(*lhs,rhs);
-	return SUCCESS;
-}
-
-/* union together two arrays, ending up with unique array (b) */
-static void
-union_arrays(a,b,acount,bcount)
-int *a, *b;
-int acount, *bcount;
-{
-	int i, j;
-
-	for (i=0; i < acount ;i++) {
-		/* check this one against all so far */
-		for (j=0;j < *bcount;j++) {
-			if (a[i] == b[j]) break;
-		}
-		/* if not found, add to array */
-		if (j== *bcount) {
-			b[j] = a[i];
-			(*bcount)++;
 		}
 	}
-}
 
 /* return TRUE if string appears to be a set of arguments
    The intent of this test is to support the ability of commands to have
@@ -361,10 +289,8 @@ char **argv;
 	char *a;
 
 	/* + 2 is for blank separating cmd and null at end */
-	if (!(buf = malloc(strlen(argv[0]) + strlen(argv[1]) + 2))) {
-		exp_error(interp,"%s: no space to save arguments",argv[0]);
-		return(TCL_ERROR);
-	}
+	buf = ckalloc(strlen(argv[0]) + strlen(argv[1]) + 2);
+
 	/* replace top-level newlines with blanks */
 	for (a=argv[1];*a;) {
 		extern char *TclWordEnd();
@@ -378,7 +304,11 @@ char **argv;
 	/* recreate statement */
 	sprintf(buf,"%s %s",argv[0],argv[1]);
 
+#if TCL_MAJOR_VERSION == 6
 	rc = Tcl_Eval(interp,buf,0,(char **)NULL);
+#else
+	rc = Tcl_Eval(interp,buf);
+#endif
 	free(buf);
 	return(rc);
 
@@ -407,55 +337,42 @@ char **argv;
 	}
 
 
+#if TCL_MAJOR_VERSION == 6
 	rc = Tcl_Eval(interp,buf,0,(char **)NULL);
+#else
+	rc = Tcl_Eval(interp,buf);
+#endif
 	free(buf);
 	return(rc);
 #endif
 }
 
-/* the following are just reserved addresses, to be used as ClientData */
-/* args to be used to tell commands how they were called. */
-/* The actual values won't be used, only the addresses, but I give them */
-/* values out of my irrational fear the compiler might collapse them all. */
-int expectCD_user	= 0;	/* called as expect_user */
-int expectCD_process	= 1;	/* called as expect */
-int expectCD_tty	= 2;	/* called as expect_tty */
-
-#define EXPECT_USER	(clientData == &expectCD_user)
-#define EXPECT_TTY	(clientData == &expectCD_tty)
-
-/* parse the arguments to expect or it's variants */
-/* returns TCL_OK or TCL_ERROR */
-static int
-parse_expect_args(clientData,interp,argc,argv,ecases,ecases_inuse,masters,mcount,me)
-ClientData clientData;
+/* parse the arguments to expect or its variants */
+/* returns TCL_ERROR for failure */
+/* (can't return i_list directly because there is no way to differentiate between */
+/* clearing, say, expect_before and signalling an error) */
+int
+parse_expect_args(interp,eg,default_spawn_id,argc,argv,ecases,ecases_inuse,i_list)
 Tcl_Interp *interp;
+struct expect_global *eg;
+int default_spawn_id;	/* suggested master if called as expect_user or expect_tty */
 int argc;
 char **argv;
 struct ecase **ecases;
 int *ecases_inuse;
-int **masters;		/* array of unique masters in ecases */
-int *mcount;		/* count of masters */
-int me;
+struct exp_i **i_list;		/* caller assumed to zero */
 {
-	int size;
 	int i;
 	struct ecase *ec;
-	int m;
-
-	int case_master;	/* master to assign to next case as it is */
-				/* being parsed */
+	int look_for_pattern;	/* indicates that keyword has not been seen */
+	int look_for_keyword;	/* accept keyword unless "--" seen */
+	int direct = TRUE;	/* indirect if false */
+	int before_or_after;	/* if called as expect_before or after */
 
 	argv++;
 	argc--;
 
-	/* if we want to listen to user, use fd 0 */
-	if (EXPECT_USER) m = 0;	/* true if we were called as expect_user */
-	else if (EXPECT_TTY) m = exp_dev_tty;	/* called as expect_tty */
-	else {
-		/* it'll be checked later, if used */
-		(void) exp_update_master(interp,&m,0,0);
-	}
+	before_or_after = !!eg;
 
 	*ecases_inuse = (1+argc)/2;	/* estimate of number of patterns */
 
@@ -464,34 +381,36 @@ int me;
 	/* trouble of making two passes, so we'll just adjust the count */
 	/* later when we find out the real amount */
 
-	if (*ecases) free((char *)*ecases);	/* for before/after cases */
 	if (*ecases_inuse) {
-		if (0 == (*ecases = (struct ecase *)malloc(*ecases_inuse *
-						sizeof(struct ecase)))) {
-			exp_error(interp,"malloc(%d ecases)",*ecases_inuse);
-			goto error;
-		}
+		*ecases = (struct ecase *)ckalloc(*ecases_inuse *
+						sizeof(struct ecase));
 	} else *ecases = NULL;
 
 	/* zero them out so that just in case we get an error in the middle */
 	/* of SplitList, we can deallocate them cleanly */
 	for (i = 0, ec = *ecases;i<argc;i+=2,ec++) {
-		ec->patn = 0;	/* necessary? - makes freeing code simpler? */
+		ec->i_list = 0;
 		ec->pat = 0;
 		ec->body = 0;
-		ec->transfer = 1;
+		ec->transfer = TRUE;
+		ec->iwrite = FALSE;
+		ec->iread = FALSE;
+		ec->timestamp = FALSE;
 		ec->re = 0;
 		ec->Case = CASE_NORM;
-		ec->owner = me;
 		ec->use = PAT_GLOB;
 	}
 
 	/* forget old estimate of cases and prepare to calculate true number */
 	*ecases_inuse = 0;
 	ec = *ecases;
-	case_master = m;
+
+	look_for_keyword = TRUE;	/* reset at bottom of loop */
+
 	for (i = 0;i<argc;i++) {
-		if (streq(argv[i],"-n")) {
+		if (streq(argv[i],"--")) {
+			look_for_keyword = FALSE;
+		} else if (streq(argv[i],"-n")) {
 			ec->transfer = 0;
 			continue;
 		} else if (streq(argv[i],"-re")) {
@@ -500,46 +419,74 @@ int me;
 		} else if (streq(argv[i],"-nocase")) {
 			ec->Case = CASE_LOWER;
 			continue;
+		} else if (streq(argv[i],"-indirect")) {
+			direct = FALSE;
+			continue;
 		} else if (streq(argv[i],"-i")) {
 			i++;
 			if (i>=argc) {
 				exp_error(interp,"-i requires following spawn_id");
 				goto error;
 			}
-			case_master = atoi(argv[i]);
+
+			ec->i_list = exp_new_i_complex(interp,/*direct,*/ argv[i],before_or_after);
+
+			if (!ec->i_list) goto error;
+
+			/* link new i_list to head of list */
+			ec->i_list->next = *i_list;
+			*i_list = ec->i_list;
+
+			direct = TRUE;
+			continue;
+		} else if (streq(argv[i],"-iwrite")) {
+			ec->iwrite = TRUE;
+			continue;
+		} else if (streq(argv[i],"-iread")) {
+			ec->iread = TRUE;
+			continue;
+		} else if (streq(argv[i],"-timestamp")) {
+			ec->timestamp = TRUE;
 			continue;
 		}
 
-		ec->m = case_master;
+		if (!ec->i_list) {
+			/* if no -i flag has occurred, use default */
+			if (!*i_list) {
+				if (default_spawn_id != EXP_SPAWN_ID_BAD) {
+					*i_list = exp_new_i_simple(interp,default_spawn_id);
+		} else {
+					/* it'll be checked later, if used */
+					(void) exp_update_master(interp,&default_spawn_id,0,0);
+					*i_list = exp_new_i_simple(interp,default_spawn_id);
+				}
+			}
+			ec->i_list = *i_list;
+		}
 
 		/* save original pattern spec */
-		if (save_str(&ec->pat,argv[i],me)
-		 || save_str(&ec->body,argv[i+1],me)) {
-			exp_error(interp,"no space to save: <%s> and <%s>",
-				argv[i],argv[i+1]);
-			goto error;
-		}
+		save_str(&ec->pat,argv[i],!before_or_after);
+		save_str(&ec->body,argv[i+1],!before_or_after);
 			
+		look_for_pattern = TRUE;
+		if (look_for_keyword) {
 		if (streq(argv[i],"timeout")) {
 			ec->use = PAT_TIMEOUT;
+				look_for_pattern = FALSE;
 		} else if (streq(argv[i],"eof")) {
 			ec->use = PAT_EOF;
+				look_for_pattern = FALSE;
 		} else if (streq(argv[i],"full_buffer")) {
 			ec->use = PAT_FULLBUFFER;
+				look_for_pattern = FALSE;
 		} else if (streq(argv[i],"default")) {
 			ec->use = PAT_DEFAULT;
-		} else if (streq(argv[i],"-brace")) {
-			/* allow exp_eval_with_one_arg to work */
-		} else if (streq(argv[i],"-dont_use")) {
-			if (TCL_OK != Tcl_SplitList(interp,argv[i],
-						&ec->n,&ec->patn)) {
-				errorlog("%s\r\n",interp->result);
-				exp_error(interp,"failed to parse pattern: %s",argv[i]);
-				goto error;
+				look_for_pattern = FALSE;
 			}
-		} else { /* pattern */
+		}
+
+		if (look_for_pattern) {
 			if (ec->use == PAT_RE) {
-			  ec->n = 1;
 			  tclRegexpError = 0;
 			  if (!(ec->re = regcomp(argv[i]))) {
 				exp_error(interp,"bad regular expression: %s",
@@ -548,63 +495,35 @@ int me;
 			  }
 		        } else {
 			  ec->use = PAT_GLOB;
-			  ec->n = 1;
-			  ec->patn = (char **)malloc(sizeof(char *));
-			  if (!ec->patn) {
-				exp_error(interp,"malloc failed while parsing pattern: %s",argv[i]);
-				goto error;
-			  }
-			  ec->patn[0] = argv[i];
 			}
 		}
+		look_for_keyword = TRUE;
 		i++; ec++; (*ecases_inuse)++;
 	}
 
-	/* build a list of masters for later use by ready() */
-	size = *ecases_inuse;
-	if (me == EXP_DURING) {
-		size += 1 + before.mcount + after.mcount;
-		/* "+1" just in case no patterns given, need placeholder */
-		/* for default pattern */
-	}
-	if (0 == (*masters = (int *)malloc(size*sizeof(int)))) {
-		exp_error(interp,"malloc(%d spawn_id's)",size);
-		goto error;
-	}
+	if (!before_or_after) {
+		/* if no patterns at all have appeared force the current spawn id */
+		/* to be added to list anyway */
 
-	*mcount = 0;	/* initially empty */
-	for (i=0,ec= *ecases; i < *ecases_inuse ;i++,ec++) {
-		int j;
-
-		if (ec->m == SPAWN_ID_ANY) continue;
-
-		/* check this one against all so far */
-		for (j=0;j < *mcount;j++) {
-			if (ec->m == (*masters)[j]) break;
+		if (*i_list == 0) {
+			if (default_spawn_id != EXP_SPAWN_ID_BAD) {
+				*i_list = exp_new_i_simple(interp,default_spawn_id);
+			} else {
+				/* it'll be checked later, if used */
+				(void) exp_update_master(interp,&default_spawn_id,0,0);
+				*i_list = exp_new_i_simple(interp,default_spawn_id);
 		}
-		/* if not found, add to array */
-		if (j== *mcount) {
-			(*masters)[j] = ec->m;
-			(*mcount)++;
 		}
 	}
 
-	/* add before/after masters to candidates to pass to ready */
-	if (me == EXP_DURING) {
-		if (*mcount == 0) {
-			/* if no patterns given, force pattern default */
-			/* with current master */
-			(*masters)[0] = m;
-			(*mcount)++;
-		}
-		union_arrays(before.masters,*masters,before.mcount,mcount);
-		union_arrays(after.masters,*masters,after.mcount,mcount);
-	}
-
 	return(TCL_OK);
 
  error:
-	if (*ecases) free_ecases(*ecases,*ecases_inuse,me);
+	/* very hard to free case_master_list here if it hasn't already */
+	/* been attached to a case, ugh */
+
+	if (*i_list) exp_free_i(*i_list);
+	if (*ecases) free_ecases(*ecases,*ecases_inuse,before_or_after);
 	*ecases = NULL;
 	*ecases_inuse = 0;
 	return(TCL_ERROR);
@@ -667,15 +586,11 @@ int *last_case;
 		        }
 		    }
 	} else if (e->use == PAT_GLOB) {
-		int j;
 		int match; /* # of chars that matched */
 
-		for (j=0;j<e->n;j++) {
-			/* skip place-holder for spawn_id with no pattern */
-			if (e->patn[j][0] == '\0') continue;
-		        debuglog("{%s}? ",dprintify(e->patn[j]));
+	        debuglog("{%s}? ",dprintify(e->pat));
 			if (buffer && (-1 != (match = Exp_StringMatch(
-					buffer,e->patn[j],&e->glob_start)))) {
+				buffer,e->pat,&e->glob_start)))) {
 				o->e = e;
 				o->match = match;
 				o->buffer = buffer;
@@ -683,7 +598,6 @@ int *last_case;
 				debuglog(yes);
 				return(EXP_MATCH);
 			} else debuglog(no);
-		}
 	} else if ((f->size == f->msize) && (f->size > 0)) {
 		debuglog("%s? ",e->pat);
 		o->e = e;
@@ -712,6 +626,7 @@ int *masters;
 int mcount;
 {
 	int i;
+	int em;	/* master of ecase */
 	struct ecase *e;
 
 	if (o->e || status == EXP_TCLERROR) return(status);
@@ -727,9 +642,14 @@ int mcount;
 	} else if (status == EXP_EOF) {
 		for (i=0, e=ecs_in;i<ecases_inuse;i++,e++) {
 			if (e->use == PAT_EOF || e->use == PAT_DEFAULT) {
-				if (e->m == SPAWN_ID_ANY || e->m == m) {
+				struct exp_fd_list *fdl;
+
+				for (fdl=e->i_list->fd_list; fdl ;fdl=fdl->next) {
+					em = fdl->fd;
+					if (em == SPAWN_ID_ANY || em == m) {
 					o->e = e;
-					break;
+						return(status);
+					}
 				}
 			}
 		}
@@ -742,39 +662,32 @@ int mcount;
 	/* The bufferful condition does not prevent a pattern match from */
 	/* occurring and vice versa, so it is scanned with patterns */
 	for (i=0, e=ecs_in;i<ecases_inuse;i++,e++) {
+		struct exp_fd_list *fdl;
 		int j;
 
 		if (e->use == PAT_TIMEOUT ||
 		    e->use == PAT_DEFAULT ||
 		    e->use == PAT_EOF) continue;
 
-#if 0
-		/* if m == SPAWN_ID_ANY, then we have not yet read from any */
-		/* master, so check every case against its master */
-#endif
-
-		/* if e->m == SPAWN_ID_ANY, then user is explicitly asking */
+		for (fdl = e->i_list->fd_list; fdl; fdl = fdl->next) {
+			em = fdl->fd;
+			/* if em == SPAWN_ID_ANY, then user is explicitly asking */
 		/* every case to be checked against every master */
-		if (e->m == SPAWN_ID_ANY) {
+			if (em == SPAWN_ID_ANY) {
 			/* test against each spawn_id */
 			for (j=0;j<mcount;j++) {
 				status = eval_case_string(interp,e,masters[j],o,last_f,last_case);
 				if (status != EXP_NOMATCH) return(status);
 			}
-#if 0
-		} else if (m == SPAWN_ID_ANY) {
-			/* test against its own spawn_id */
-			status = eval_case_string(interp,e,e->m,o,last_f,last_case);
-			if (status != EXP_NOMATCH) return(status);
-#endif
 		} else {
 			/* reject things immediately from wrong spawn_id */
-			if (e->m != m) continue;
+				if (em != m) continue;
 
 			status = eval_case_string(interp,e,m,o,last_f,last_case);
 			if (status != EXP_NOMATCH) return(status);
 		}
 	}
+	}
 	return(EXP_NOMATCH);
 }
 
@@ -788,15 +701,25 @@ Tcl_Interp *interp;
 int argc;
 char **argv;
 {
-	struct expect_special *e;
+	struct expect_global *eg;
+	eg = (struct expect_global *) clientData;
 
-	if (argc == 2 && strchr(argv[1],'\n')) {
+	if (eg->i_list) exp_free_i(eg->i_list);
+	eg->i_list = 0;
+	if (eg->ecases) free_ecases(eg->ecases,eg->ecount,1);
+	eg->ecases = 0;
+
+	if ((argc == 2) && exp_one_arg_braced(argv[1])) {
 		return(exp_eval_with_one_arg(clientData,interp,argc,argv));
+	} else if ((argc == 3) && streq(argv[1],"-brace")) {
+		char *new_argv[2];
+		new_argv[0] = argv[0];
+		new_argv[1] = argv[2];
+		return(exp_eval_with_one_arg(clientData,interp,argc,new_argv));
 	}
 
-	e = (struct expect_special *) clientData;
-	return(parse_expect_args(clientData,interp,argc,argv,&e->ecases,
-		&e->ecount,&e->masters,&e->mcount,e->me));
+	return(parse_expect_args(interp,eg,EXP_SPAWN_ID_BAD,argc,argv,&eg->ecases,
+		&eg->ecount,&eg->i_list));
 }
 
 /* adjusts file according to user's size request */
@@ -807,7 +730,6 @@ Tcl_Interp *interp;
 struct f *f;
 {
 	int new_msize;
-	char *new_buf;
 
 	/* get the latest buffer size.  Double the user input for */
 	/* two reasons.  1) Need twice the space in case the match */
@@ -821,29 +743,13 @@ struct f *f;
 	if (new_msize != f->msize) {
 		if (!f->buffer) {
 			/* allocate buffer space for 1st time */
-			f->lower = malloc((unsigned)new_msize+1);
-			f->buffer = malloc((unsigned)new_msize+1);
-			if ((!f->buffer) && (!f->lower)) {
-				exp_error(interp,"out of space - failed to malloc initial match buffer");
-				if (f->lower) {free(f->lower); f->lower = 0;}
-				return(TCL_ERROR);
-			}
+			f->lower = ckalloc((unsigned)new_msize+1);
+			f->buffer = ckalloc((unsigned)new_msize+1);
 			f->size = 0;
 		} else {
 			/* buffer already exists - resize */
-			if (0 == (new_buf = realloc(f->buffer,new_msize+1))) {
-				exp_error(interp,"failed to grow match buf to %d bytes",f->umsize);
-				return(TCL_ERROR);
-			}
-			f->buffer = new_buf;
-			if (0 == (new_buf = realloc(f->lower,new_msize+1))) {
-				exp_error(interp,"failed to grow match buf to %d bytes",f->umsize);
-				/* no need to free other buffer */
-				/* - code will still work even if */
-				/* buffer is larger than necessary */
-				return(TCL_ERROR);
-			}
-			f->lower = new_buf;
+			f->buffer = realloc(f->buffer,new_msize+1);
+			f->lower = realloc(f->lower,new_msize+1);
 			/* if truncated, forget about some data */
 			if (f->size >= f->msize) f->size = f->msize-1;
 		}
@@ -950,7 +856,8 @@ int key;
 		}
 		/* don't write to user if they're seeing it already, */
 		/* that is, typing it! */
-		if (loguser && !f_is_user(f)) fwrite(f->buffer + f->printed,
+		if (loguser && !exp_is_stdinfd(*m))
+			fwrite(f->buffer + f->printed,
 					1,write_count,stdout);
 		if (debugfile) fwrite(f->buffer + f->printed,
 					1,write_count,debugfile);
@@ -1005,7 +912,7 @@ int timeout;
 			if (f->printed < 0) f->printed = 0;
 		}
 
-#if SIMPLE_EVENT
+#ifdef SIMPLE_EVENT
 		alarm((timeout > 0)?timeout:1);
 #endif
 
@@ -1019,7 +926,7 @@ debuglog("= %d\r\n",i_read_cc);
 #endif
 		i_read_errno = errno;	/* errno can be overwritten by the */
 					/* time we return */
-#if SIMPLE_EVENT
+#ifdef SIMPLE_EVENT
 		alarm(0);
 #endif
 	}
@@ -1081,6 +988,44 @@ void exp_disarm_event_handlers();
 #endif
 #endif
 
+int
+update_expect_fds(interp,i_list,fd_union)
+Tcl_Interp *interp;
+struct exp_i *i_list;
+struct exp_fd_list **fd_union;
+{
+	struct exp_i *p;
+
+	/* for each i_list in an expect statement ... */
+	for (p=i_list;p;p=p->next) {
+		struct exp_fd_list *fdl;
+
+		/* do not update "direct" entries (again) */
+		/* they were updated upon creation */
+		if (!p->direct) {
+			if (TCL_OK != exp_i_update(interp,p)) return TCL_ERROR;
+		}
+
+		/* for each fd in the i_list */
+		for (fdl=p->fd_list;fdl;fdl=fdl->next) {
+			struct exp_fd_list *u;
+
+			if (fdl->fd == SPAWN_ID_ANY) continue;
+
+			/* check this one against all so far */
+			for (u = *fd_union;u;u=u->next) {
+				if (fdl->fd == u->fd) goto found;
+			}
+			/* if not found, link in as head of list */
+			fdl = exp_new_fd(interp,fdl->fd);
+			fdl->next = *fd_union;
+			*fd_union = fdl;
+		found:;
+		}
+	}
+	return TCL_OK;
+}
+
 /*ARGSUSED*/
 int
 cmdExpect(clientData, interp, argc, argv)
@@ -1098,12 +1043,15 @@ char **argv;
 	int i;			/* trusty temporary */
 	struct ecase *ecases;
 	int ecases_inuse;	/* number of ecases to use */
+	struct exp_fd_list *fd_list;	/* list of masters to watch */
+	struct exp_fd_list *fdl;	/* temp for interating over fd_list */
 	int *masters;		/* array of masters to watch */
 	int mcount;		/* number of masters to watch */
 
 	struct eval_out eo;	/* final case of interest */
 
 	int result;		/* Tcl result */
+	int iread;		/* if true, reread indirects */
 
 	time_t start_time_total;/* time at beginning of this procedure */
 	time_t start_time = 0;	/* time when restart label hit */
@@ -1119,14 +1067,69 @@ char **argv;
 
 	int key;		/* identify this expect command instance */
 
+	int close_count;	/* monitor exp_close_count */
+
 	int remtime;		/* remaining time in timeout */
 
+	/* list corresponding to set of -i flags */
+	struct exp_i *during_i_list;
+
 	if ((argc == 2) && exp_one_arg_braced(argv[1])) {
 		return(exp_eval_with_one_arg(clientData,interp,argc,argv));
+	} else if ((argc == 3) && streq(argv[1],"-brace")) {
+		char *new_argv[2];
+		new_argv[0] = argv[0];
+		new_argv[1] = argv[2];
+		return(exp_eval_with_one_arg(clientData,interp,argc,new_argv));
 	}
 
 	time(&start_time_total);
 	start_time = start_time_total;
+
+	/* make arg list for processing cases */
+	/* do it dynamically, since expect can be called recursively */
+
+	during_i_list = 0;
+	fd_list = 0;
+	ecases = 0;
+	masters = 0;
+	if (TCL_ERROR == parse_expect_args(interp,(struct expect_global *)0,
+			(int)clientData,argc,argv,
+			&ecases,&ecases_inuse,&during_i_list)) return TCL_ERROR;
+
+ restart_with_update:
+	/* get the latest timeout */
+	(void) get_timeout(interp);
+
+	/* validate all descriptors */
+	/* and flatten fds into array */
+
+	if ((TCL_ERROR == update_expect_fds(interp,before.i_list,&fd_list))
+	 || (TCL_ERROR == update_expect_fds(interp,after.i_list, &fd_list))
+	 || (TCL_ERROR == update_expect_fds(interp,during_i_list,&fd_list))) {
+		result = TCL_ERROR;
+		goto cleanup;
+	}
+
+	close_count = exp_close_count;
+
+	/* count and validate fd_list */
+	mcount = 0;
+	for (fdl=fd_list;fdl;fdl=fdl->next) {
+		mcount++;
+		/* validate all input descriptors */
+		if (!exp_fd2f(interp,fdl->fd,1,1,"expect")) {
+			result = TCL_ERROR;
+			goto cleanup;
+		}
+	}
+
+	/* make into an array */
+	masters = (int *)ckalloc(mcount * sizeof(int));
+	for (fdl=fd_list,i=0;fdl;fdl=fdl->next,i++) {
+		masters[i] = fdl->fd;
+	}
+
  restart:
 	if (first_time) first_time = 0;
 	else time(&start_time);
@@ -1137,35 +1140,31 @@ char **argv;
 #if 0
 	m = SPAWN_ID_ANY;
 #endif
-	ecases = 0;
-	masters = 0;
-	mcount = 0;
 	result = TCL_OK;
 	last_f = 0;
+
+	/* if we've done a "continue -expect" without going rereading and */
+	/* rechecking all the indirects, it's still possible that some fds */
+	/* have been closed.  Check the close_count and if different */
+	/* verify all the descriptors */
+	if (close_count != exp_close_count) {
+		close_count = exp_close_count;
+
+		for (fdl=fd_list;fdl;fdl=fdl->next) {
+		/* validate all input descriptors */
+			if (!exp_fd2f(interp,fdl->fd,1,1,"expect")) {
+			result = TCL_ERROR;
+			goto cleanup;
+		}
+	}
+	}
+
 	/* end of restart code */
 
 	eo.e = 0;		/* no final case yet */
 	eo.f = 0;		/* no final file selected yet */
 	eo.match = 0;		/* nothing matched yet */
 
-	/* get the latest timeout */
-	(void) get_timeout(interp);
-
-	/* make arg list for processing cases */
-	/* do it dynamically, since expect can be called recursively */
-	if (TCL_ERROR == parse_expect_args(clientData,interp,argc,argv,
-			&ecases,&ecases_inuse,
-			&masters,&mcount,
-			EXP_DURING)) return(TCL_ERROR);
-
-	for (i=0;i<mcount;i++) {
-		/* validate all input descriptors */
-		if (!(f = exp_fd2f(interp,masters[i],1,1,"expect"))) {
-			result = TCL_ERROR;
-			goto cleanup;
-		}
-	}
-
 	/* timeout code is a little tricky, be very careful changing it */
 	if (timeout != EXP_TIME_INFINITY) {
 		time(&current_time);
@@ -1241,21 +1240,12 @@ error:
  done:
 #define out(i,val)  debuglog("expect: set %s(%s) {%s}\r\n",EXPECT_OUT,i, \
 						dprintify(val)); \
-		    if (!Tcl_SetVar2(interp,EXPECT_OUT,i,val,0)) \
-				{result = TCL_ERROR; goto cleanup;}
-	{
-		char value[20];
+		    Tcl_SetVar2(interp,EXPECT_OUT,i,val,0);
 
-		time(&current_time);
-		elapsed_time = current_time - start_time;
-		elapsed_time_total = current_time - start_time_total;
-		sprintf(value,"%d",elapsed_time);
-		out("seconds",value);
-		sprintf(value,"%d",elapsed_time_total);
-		out("seconds_total",value);
-	}
+	iread = FALSE;
 
 	if (result != TCL_ERROR) {
+		int iwrite = FALSE;	/* write spawn_id? */
 		char *body = 0;
 		char *buffer;	/* pointer to normal or lowercased data */
 		struct ecase *e = 0;	/* points to current ecase */
@@ -1266,12 +1256,26 @@ error:
 		if (eo.e) {
 			e = eo.e;
 			body = e->body;
+			iread = e->iread;
+			iwrite = e->iwrite;
 			if (cc != EXP_TIMEOUT) {
-/*			if (e->use != PAT_TIMEOUT) {*/
 				f = eo.f;
 				match = eo.match;
 				buffer = eo.buffer;
 			}
+			if (e->timestamp) {
+				char value[20];
+
+				time(&current_time);
+				elapsed_time = current_time - start_time;
+				elapsed_time_total = current_time - start_time_total;
+				sprintf(value,"%d",elapsed_time);
+				out("seconds",value);
+				sprintf(value,"%d",elapsed_time_total);
+				out("seconds_total",value);
+
+				exp_timestamp(interp,&current_time,EXPECT_OUT);
+			}
 		} else if (cc == EXP_EOF) {
 			/* read an eof but no user-supplied case */
 			f = eo.f;
@@ -1286,7 +1290,7 @@ error:
 				for (i=0;i<NSUBEXP;i++) {
 					int offset;
 
-					if (e->re->startp[i] == 0) break;
+					if (e->re->startp[i] == 0) continue;
 
 					/* start index */
 					sprintf(name,"%d,start",i);
@@ -1302,10 +1306,7 @@ error:
 
 					/* string itself */
 					sprintf(name,"%d,string",i);
-#if 0
-	/* doesn't seem to be used? */
-					str = f->buffer + offset;*/
-#endif
+
 					/* temporarily null-terminate in */
 					/* middle */
 					match_char = *e->re->endp[i];
@@ -1348,8 +1349,10 @@ error:
 		if (eo.f) {
 			char spawn_id[10];	/* enough for a %d */
 
+			if (iwrite) {
 			sprintf(spawn_id,"%d",f-fs);
 			out("spawn_id",spawn_id);
+			}
 
 			/* save buf[0..match] */
 			/* temporarily null-terminate string in middle */
@@ -1377,21 +1380,84 @@ error:
 		}
 
 		if (body) {
+#if TCL_MAJOR_VERSION == 6
 			result = Tcl_Eval(interp,body,0,(char **) NULL);
+#else
+			result = Tcl_Eval(interp,body);
+#endif
 		}
 	}
 
  cleanup:
-	free_ecases(ecases,ecases_inuse,EXP_DURING);
-	if (masters) free((char *)masters);
-
-	if (result == TCL_CONTINUE_EXPECT) {
+	if ((result == TCL_CONTINUE_EXPECT) && !iread) {
 		debuglog("expect: continuing expect\r\n");
 		goto restart;
 	}
 
+	if (fd_list) {
+		exp_free_fd(fd_list);
+		fd_list = 0;
+	}
+	if (masters) {
+		free((char *)masters);
+		masters = 0;
+	}
+
+	if (result == TCL_CONTINUE_EXPECT) {
+		debuglog("expect: continuing expect after update\r\n");
+		goto restart_with_update;
+	}
+
+	exp_free_i(during_i_list);
+
+	free_ecases(ecases,ecases_inuse,0);
+
 	return(result);
 }
+#undef out
+
+
+#define out(elt) \
+    debuglog("expect: set %s(%s) {%s}\r\n",array,elt,dprintify(ascii)); \
+    Tcl_SetVar2(interp,array,elt,ascii,0);
+
+void
+exp_timestamp(interp,timeval,array)
+Tcl_Interp *interp;
+time_t *timeval;
+char *array;
+{
+	struct tm *tm;
+	char *ascii;
+
+	tm = localtime(timeval);	/* split */
+	ascii = asctime(tm);		/* print */
+	ascii[24] = '\0';		/* zap trailing \n */
+
+	out("timestamp");
+
+	sprintf(ascii,"%ld",*timeval);
+	out("epoch");
+
+	sprintf(ascii,"%d",tm->tm_sec);
+	out("sec");
+	sprintf(ascii,"%d",tm->tm_min);
+	out("min");
+	sprintf(ascii,"%d",tm->tm_hour);
+	out("hour");
+	sprintf(ascii,"%d",tm->tm_mday);
+	out("mday");
+	sprintf(ascii,"%d",tm->tm_mon);
+	out("mon");
+	sprintf(ascii,"%d",tm->tm_year);
+	out("year");
+	sprintf(ascii,"%d",tm->tm_wday);
+	out("wday");
+	sprintf(ascii,"%d",tm->tm_yday);
+	out("yday");
+	sprintf(ascii,"%d",tm->tm_isdst);
+	out("isdst");
+}
 
 /* lowmemcpy - like memcpy but it lowercases result */
 void
@@ -1641,9 +1707,6 @@ exp_init_trap()
 #if defined(SIGALRM)		/* alarm clock */
 	signals[SIGALRM].name = "*SIGALRM";
 #endif
-#if defined(SIGCLD)		/* Like SIGCHLD.  */
-	signals[SIGCLD ].name = "SIGCLD";
-#endif
 #if defined(SIGPWR)		/* imminent power failure */
 	signals[SIGPWR ].name = "SIGPWR";
 #endif
@@ -1664,6 +1727,10 @@ exp_init_trap()
 #endif
 #if defined(SIGCHLD)		/* to parent on child stop or exit */
 	signals[SIGCHLD].name = "SIGCHLD";
+#else
+#if defined(SIGCLD)		/* Like SIGCHLD.  */
+	signals[SIGCLD ].name = "SIGCHLD";
+#endif
 #endif
 #if defined(SIGTTIN)		/* to readers pgrp upon background tty read */
 	signals[SIGTTIN].name = "SIGTTIN";
@@ -1791,11 +1858,11 @@ int sig;
 			for (i=0;i<=fd_max;i++) {
 				if (fs[i].pid == pid) break;
 			}
-			if (i>fd_max || !(fs[i].flags & FD_VALID)) {
+			if (i>fd_max || !(fs[i].valid)) {
 				debuglog("received SIGCLD from unknown pid %d\r\n",pid);
 				proc_valid = FALSE;
 			} else {
-				fs[i].flags |= FD_SYSWAITED;
+				fs[i].sys_waited = TRUE;
 				fs[i].wait = status;
 			}
 		}
@@ -1807,7 +1874,11 @@ int sig;
 		if (proc_valid) {
 		  Tcl_Interp *interp = signals[sig].interp;
 
+#if TCL_MAJOR_VERSION == 6
 		  rc = Tcl_Eval(interp,signals[sig].action,0,(char **)0);
+#else
+		  rc = Tcl_Eval(interp,signals[sig].action);
+#endif
 		  if (rc != TCL_OK) {
 		    errorlog("caught %s (%d): error in command: %s\r\n",
 			signal_to_string(sig),sig,signals[sig].action);
@@ -1832,6 +1903,8 @@ int sig;
 
 	signals[sig].action = 0;  /* should've been free'd by now if nec. */
 
+	if (sig == 0) return;	
+
 	/* SIGINT defaults to timeout/exit routine */
 	/* Ultrix 1.3 compiler can't handle this */
 	/* default_proc = (sig == SIGINT?sigint_handler:SIG_DFL);*/
@@ -2045,18 +2118,20 @@ char **argv;
 
 		debuglog("trap: setting up signal %d (\"%s\")\r\n",sig,list[i]);
 
+#ifdef TCL_DEBUGGER
+              if (sig == SIGINT && exp_tcl_debugger_available) {
+                      debuglog("trap: cannot trap SIGINT while using debugger\r\n");
+                      continue;
+              }
+#endif /* TCL_DEBUGGER */
+
 		if (signals[sig].action) free(signals[sig].action);
 
 		if (streq(action,"SIG_DFL")) {
-			if (sig != 0) sig_reset(sig);
+			sig_reset(sig);
 		} else {
 			len = 1 + strlen(action);
-			if (0 == (signals[sig].action = malloc(len))) {
-				exp_error(interp,"trap: malloc failed");
-				if (sig != 0) sig_reset(sig);
-				rc = TCL_ERROR;
-				break;
-			}
+			signals[sig].action = ckalloc(len);
 			memcpy(signals[sig].action,action,len);
 			signals[sig].interp = interp;
 			if (sig == 0) continue;
@@ -2074,21 +2149,27 @@ exp_init_expect(interp)
 Tcl_Interp *interp;
 {
 	Tcl_CreateCommand(interp,"expect",
-		cmdExpect,(ClientData)&expectCD_process,exp_deleteProc);
+		cmdExpect,(ClientData)EXP_SPAWN_ID_BAD,exp_deleteProc);
 	Tcl_CreateCommand(interp,"expect_after",
 		cmdExpectGlobal,(ClientData)&after,exp_deleteProc);
 	Tcl_CreateCommand(interp,"expect_before",
 		cmdExpectGlobal,(ClientData)&before,exp_deleteProc);
 	Tcl_CreateCommand(interp,"expect_user",
-		cmdExpect,(ClientData)&expectCD_user,exp_deleteProc);
+		cmdExpect,(ClientData)0,exp_deleteProc);
 	Tcl_CreateCommand(interp,"expect_tty",
-		cmdExpect,(ClientData)&expectCD_tty,exp_deleteProc);
+		cmdExpect,(ClientData)exp_dev_tty,exp_deleteProc);
 	Tcl_CreateCommand(interp,"match_max",
 		cmdMatchMax,(ClientData)0,exp_deleteProc);
+	Tcl_CreateCommand(interp,"exp_match_max",
+		cmdMatchMax,(ClientData)0,exp_deleteProc);
 	Tcl_CreateCommand(interp,"parity",
 		cmdParity,(ClientData)0,exp_deleteProc);
+	Tcl_CreateCommand(interp,"exp_parity",
+		cmdParity,(ClientData)0,exp_deleteProc);
 	Tcl_CreateCommand(interp,"trap",
 		cmdTrap,(ClientData)0,exp_deleteProc);
+	Tcl_CreateCommand(interp,"exp_trap",
+		cmdTrap,(ClientData)0,exp_deleteProc);
 
 	Tcl_SetVar(interp,EXPECT_TIMEOUT,	INIT_EXPECT_TIMEOUT,0);
 	Tcl_SetVar(interp,SPAWN_ID_ANY_VARNAME,	SPAWN_ID_ANY_LIT,0);
